# Copyright 2023 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for seqio.utils."""
import dataclasses
import functools
from typing import Mapping, Optional, Sequence

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
from seqio import test_utils
from seqio import utils
from seqio import vocabularies
import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds


tf.compat.v1.enable_eager_execution()

mock = absltest.mock
assert_dataset = test_utils.assert_dataset


_GLOBAL_INCREMENT: int = 2


class AnyArg(object):

  def __eq__(self, var):
    return True


class LazyTfdsLoaderTest(absltest.TestCase):

  def setUp(self):
    utils.LazyTfdsLoader._MEMOIZED_BUILDERS = {}
    super().setUp()

  def test_str(self):
    loader = utils.LazyTfdsLoader(
        name="a/b:1.0.0", data_dir="/data", split_map={"x": "y"}
    )
    self.assertEqual(
        str(loader),
        "LazyTfdsLoader(name=a/b:1.0.0, data_dir=/data)",
    )

  def test_repr(self):
    loader = utils.LazyTfdsLoader(
        name="a/b:1.0.0", data_dir="/data", split_map={"x": "y"}
    )
    self.assertEqual(
        loader.__repr__(),
        """LazyTfdsLoader(name=a/b:1.0.0, data_dir=/data, split_map={'x': 'y'}, decoders=None)""",
    )

  def test_no_tfds_version(self):
    with self.assertRaisesWithLiteralMatch(
        ValueError, "TFDS name must contain a version number, got: fake"
    ):
      utils.LazyTfdsLoader(name="fake")

  @mock.patch("tensorflow_datasets.builder")
  def test_builder_memoization(self, mock_tfds_builder):
    mock_tfds_builder.side_effect = lambda name, data_dir: ",".join(
        [name, data_dir or ""]
    )

    ds1 = utils.LazyTfdsLoader("ds1:1.0.0")
    self.assertEqual("ds1:1.0.0,", ds1.builder)
    self.assertEqual(1, tfds.builder.call_count)

    # Builder should be cached with same name.
    self.assertEqual("ds1:1.0.0,", ds1.builder)
    self.assertEqual(1, tfds.builder.call_count)

    # Same name but different data dir is a cache miss.
    ds1_dir1 = utils.LazyTfdsLoader("ds1:1.0.0", "dir1")
    self.assertEqual("ds1:1.0.0,dir1", ds1_dir1.builder)
    self.assertEqual(2, tfds.builder.call_count)
    # Same name and data dir is a cache hit.
    self.assertEqual("ds1:1.0.0,dir1", ds1_dir1.builder)
    self.assertEqual(2, tfds.builder.call_count)

    # Different name is a cache miss.
    ds2 = utils.LazyTfdsLoader("ds2:1.0.0")
    self.assertEqual("ds2:1.0.0,", ds2.builder)
    self.assertEqual(3, tfds.builder.call_count)

    # Different split map name is a cache hit.
    ds2 = utils.LazyTfdsLoader("ds2:1.0.0", split_map={"train": "validation"})
    self.assertEqual("ds2:1.0.0,", ds2.builder)
    self.assertEqual(3, tfds.builder.call_count)

    # Try calling everything again, order shouldn't matter.
    self.assertEqual("ds1:1.0.0,", ds1.builder)
    self.assertEqual("ds1:1.0.0,dir1", ds1_dir1.builder)
    self.assertEqual("ds2:1.0.0,", ds2.builder)
    self.assertEqual(3, tfds.builder.call_count)

  @mock.patch("tensorflow_datasets.load")
  def test_split_map(self, mock_tfds_load):
    seed = 0
    utils.LazyTfdsLoader._MEMOIZED_BUILDERS[("ds/c1:1.0.0", None)] = mock.Mock(
        info=mock.Mock(
            splits={
                "validation": mock.Mock(
                    num_examples=420, file_instructions=["f1", "f2"]
                ),
                "test": mock.Mock(num_examples=42, file_instructions=["f3"]),
            }
        )
    )

    ds = utils.LazyTfdsLoader(
        "ds/c1:1.0.0", split_map={"train": "validation", "validation": "test"}
    )

    # test .load()
    ds.load("train", shuffle_files=False, seed=seed)
    mock_tfds_load.assert_called_once_with(
        "ds/c1:1.0.0",
        split="validation",
        data_dir=None,
        shuffle_files=False,
        download=True,
        try_gcs=True,
        read_config=AnyArg(),
        decoders=None,
    )

    # test .size()
    self.assertEqual(420, ds.size(split="train"))
    self.assertEqual(42, ds.size(split="validation"))
    with self.assertRaises(KeyError):
      ds.size(split="test")

    # test .files()
    self.assertListEqual(["f1", "f2"], ds.files(split="train"))
    self.assertListEqual(["f3"], ds.files(split="validation"))
    with self.assertRaises(KeyError):
      ds.files(split="test")

  @mock.patch("tensorflow_datasets.load")
  def test_read_config_override_default(self, mock_tfds_load):
    ds = utils.LazyTfdsLoader(
        "ds/c1:1.0.0", split_map={"train": "validation", "validation": "test"}
    )
    ds.load("train", shuffle_files=False, seed=42)
    mock_tfds_load.assert_called_once()
    kwargs = mock_tfds_load.call_args.kwargs
    read_config = kwargs["read_config"]
    self.assertTrue(read_config.skip_prefetch)
    self.assertEqual(42, read_config.shuffle_seed)
    self.assertIsNone(read_config.shuffle_reshuffle_each_iteration)

  @mock.patch("tensorflow_datasets.load")
  def test_read_config_override(self, mock_tfds_load):
    read_config = tfds.ReadConfig()
    read_config.shuffle_reshuffle_each_iteration = True
    utils.set_tfds_read_config_override(read_config)
    ds = utils.LazyTfdsLoader(
        "ds/c1:1.0.0", split_map={"train": "validation", "validation": "test"}
    )
    ds.load("train", shuffle_files=False, seed=42)
    mock_tfds_load.assert_called_once()
    kwargs = mock_tfds_load.call_args.kwargs
    read_config = kwargs["read_config"]
    self.assertTrue(read_config.skip_prefetch)
    self.assertEqual(42, read_config.shuffle_seed)
    self.assertTrue(read_config.shuffle_reshuffle_each_iteration)
    # reset to default global override
    utils.set_tfds_read_config_override(None)



class TransformUtilsTest(parameterized.TestCase):

  def test_add_kwargs_to_transform_callable(self):
    def fn(x, y):
      return x * y

    fn = utils.add_kwargs_to_transform(fn, y=2, z=10)
    self.assertEqual(6, fn(3))

  def test_add_kwargs_to_transform_dataclass(self):
    @dataclasses.dataclass
    class Fn:
      factor: int
      y: Optional[int] = None

      def __call__(self, x):
        return self.factor * self.y * x

    fn = Fn(10)
    fn = utils.add_kwargs_to_transform(fn, y=2, z=10)
    self.assertEqual(60, fn(3))


class MapOverDatasetTest(parameterized.TestCase):

  def test_map_fn_simple(self):
    @utils.map_over_dataset
    def fn(ex):
      ex["field"] += 1
      return ex

    ds = tf.data.Dataset.from_tensor_slices({"field": range(10)})
    mapped_ds = fn(ds)
    expected_ds = [{"field": i + 1} for i in range(10)]
    self.assertListEqual(list(mapped_ds.as_numpy_iterator()), expected_ds)

  def test_map_fn_with_kwargs(self):
    @utils.map_over_dataset
    def fn(ex, val):
      ex["field"] += val
      return ex

    ds = tf.data.Dataset.from_tensor_slices({"field": range(10)})
    mapped_ds = functools.partial(fn, val=2)(ds)
    expected_ds = [{"field": i + 2} for i in range(10)]
    self.assertListEqual(list(mapped_ds.as_numpy_iterator()), expected_ds)

  def test_map_fn_with_special_kwargs(self):
    @utils.map_over_dataset
    def fn(ex, val, sequence_length):
      for key in sequence_length:
        ex[key] += val
      return ex

    ds = tf.data.Dataset.from_tensor_slices({"field": range(10)})
    # Special kwargs are configured when the preprocessor is called. Other
    # kwargs are configured when the preprocessor is created. Imitate this in
    # the following invocation.
    mapped_ds = functools.partial(fn, val=2)(ds, sequence_length={"field": -1})
    expected_ds = [{"field": i + 2} for i in range(10)]
    self.assertListEqual(list(mapped_ds.as_numpy_iterator()), expected_ds)

  def test_random_map_fn_simple(self):
    @utils.map_over_dataset(num_seeds=1)
    def fn(ex, seed):
      rand_int = tf.random.stateless_uniform([], seed, 0, 10, tf.int32)
      ex["field"] += rand_int
      return ex

    ds = tf.data.Dataset.from_tensor_slices({"field": range(10)})
    for _ in range(3):  # reproducible with fixed initial seed
      with utils.map_seed_manager(initial_seed=123):
        mapped_ds = fn(ds)  # pylint: disable=no-value-for-parameter
      results = [7, 5, 6, 6, 7, 11, 12, 16, 15, 15]
      expected_ds = [{"field": results[i]} for i in range(10)]
      print("gaurav", list(mapped_ds.as_numpy_iterator()))
      self.assertListEqual(list(mapped_ds.as_numpy_iterator()), expected_ds)

  def test_random_map_fn_with_kwargs(self):
    @utils.map_over_dataset(num_seeds=1)
    def fn(ex, seed, val):
      rand_int = tf.random.stateless_uniform([], seed, 0, 10, tf.int32)
      ex["field"] += rand_int + val
      return ex

    ds = tf.data.Dataset.from_tensor_slices({"field": range(10)})
    for _ in range(3):  # reproducible with fixed initial seed
      with utils.map_seed_manager(initial_seed=123):
        mapped_ds = functools.partial(fn, val=1)(ds)  # pylint: disable=no-value-for-parameter
      results = [8, 6, 7, 7, 8, 12, 13, 17, 16, 16]
      expected_ds = [{"field": results[i]} for i in range(10)]
      self.assertListEqual(list(mapped_ds.as_numpy_iterator()), expected_ds)

  def test_random_map_fn_with_special_kwargs(self):
    @utils.map_over_dataset(num_seeds=1)
    def fn(ex, seed, val, sequence_length):
      for key in sequence_length:
        rand_int = tf.random.stateless_uniform([], seed, 0, 10, tf.int32)
        ex[key] += rand_int + val
      return ex

    ds = tf.data.Dataset.from_tensor_slices({"field": range(10)})
    for _ in range(3):  # reproducible with fixed initial seed
      with utils.map_seed_manager(initial_seed=123):
        # Special kwargs are configured when the preprocessor is called. Other
        # kwargs are configured when the preprocessor is created. Imitate this
        # in the following invocation.
        map_fn = functools.partial(fn, val=1)
        mapped_ds = map_fn(ds, sequence_length={"field": -1})  # pylint: disable=no-value-for-parameter
      results = [8, 6, 7, 7, 8, 12, 13, 17, 16, 16]
      expected_ds = [{"field": results[i]} for i in range(10)]
      self.assertListEqual(list(mapped_ds.as_numpy_iterator()), expected_ds)

  def test_multi_seed_random_map_fn_special_kwargs(self):
    @utils.map_over_dataset(num_seeds=2)
    def fn(ex, seeds, val, sequence_length):
      for key in sequence_length:
        rand_int_1 = tf.random.stateless_uniform([], seeds[0], 0, 10, tf.int32)
        rand_int_2 = tf.random.stateless_uniform([], seeds[1], 0, 10, tf.int32)
        ex[key] += rand_int_1 + rand_int_2 + val
      return ex

    ds = tf.data.Dataset.from_tensor_slices({"field": range(10)})
    for _ in range(3):  # reproducible with fixed initial seed
      with utils.map_seed_manager(initial_seed=123):
        # Special kwargs are configured when the preprocessor is called. Other
        # kwargs are configured when the preprocessor is created. Imitate this
        # in the following invocation.
        map_fn = functools.partial(fn, val=1)
        mapped_ds = map_fn(ds, sequence_length={"field": -1})  # pylint: disable=no-value-for-parameter
      results = [13, 15, 16, 13, 9, 16, 15, 26, 18, 16]
      expected_ds = [{"field": results[i]} for i in range(10)]
      print("gaurav", list(mapped_ds.as_numpy_iterator()))
      self.assertListEqual(list(mapped_ds.as_numpy_iterator()), expected_ds)


class UtilsTest(parameterized.TestCase, tf.test.TestCase):
  _tfdict = {
      "bool": tf.constant([True, False], dtype=tf.bool),
      "int32": tf.constant([1], dtype=tf.int32),
      "int64": tf.constant([1], dtype=tf.int64),
      "float": tf.constant([1.0]),
      "string": tf.constant(["a"]),
      "2d_tensor": tf.reshape(tf.range(4), [2, 2]),
      "3d_tensor": tf.reshape(tf.range(6), [2, 1, 3]),
      "2d_ragged": tf.ragged.constant([[1], [2, 3]]),
      "3d_ragged": tf.ragged.constant([[[1]], [[2, 3], [4, 5, 6]]]),
      "2d_sparse": tf.sparse.SparseTensor(
          indices=[[0, 1], [2, 3]],
          values=[10, 20],
          dense_shape=[3, 4],
      ),
      "3d_sparse": tf.sparse.SparseTensor(
          indices=[[0, 1, 2], [3, 4, 5]],
          values=[10, 20],
          dense_shape=[4, 5, 6],
      ),
  }
  _tfexample = tf.train.Example(
      features=tf.train.Features(
          feature={
              "bool": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1, 0])
              ),
              # NOTE: TFExamples only stores int64s, so we can't avoid the
              # up-casting here.
              "int32": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1])
              ),
              "int64": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1])
              ),
              "float": tf.train.Feature(
                  float_list=tf.train.FloatList(value=[1.0])
              ),
              "string": tf.train.Feature(
                  bytes_list=tf.train.BytesList(value=[b"a"])
              ),
              "2d_tensor": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[0, 1, 2, 3])
              ),
              "_sh:2d_tensor": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[2, 2])
              ),
              "3d_tensor": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[0, 1, 2, 3, 4, 5])
              ),
              "_sh:3d_tensor": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[2, 1, 3])
              ),
              "2d_ragged": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1, 2, 3])
              ),
              "_rl:0:2d_ragged": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1, 2])
              ),
              "3d_ragged": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1, 2, 3, 4, 5, 6])
              ),
              "_rl:0:3d_ragged": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1, 2])
              ),
              "_rl:1:3d_ragged": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1, 2, 3])
              ),
              "2d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[10, 20])
              ),
              "_sp:0:2d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[0, 2])
              ),
              "_sp:1:2d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1, 3])
              ),
              "_sh:2d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[3, 4])
              ),
              "3d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[10, 20])
              ),
              "_sp:0:3d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[0, 3])
              ),
              "_sp:1:3d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[1, 4])
              ),
              "_sp:2:3d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[2, 5])
              ),
              "_sh:3d_sparse": tf.train.Feature(
                  int64_list=tf.train.Int64List(value=[4, 5, 6])
              ),
          }
      )
  )

  def assertTensorDictEqual(
      self,
      expected: utils.NestedTFDict,
      actual: utils.NestedTFDict,
  ):
    self.assertEqual(expected.keys(), actual.keys())
    for key, expected_value in expected.items():
      if isinstance(expected_value, dict):
        # Default assertEqual implementation does not check equality of nested
        # dictionaries.
        self.assertTensorDictEqual(expected_value, actual[key])
      elif isinstance(expected_value, tf.SparseTensor):
        # Default assertEqual implementation does not check equality of sparse
        # tensors.
        self.assertAllEqual(
            tf.sparse.to_dense(expected_value), tf.sparse.to_dense(actual[key])
        )
      else:
        self.assertAllEqual(expected_value, actual[key])

  def test_dict_to_tfexample(self):
    expected = self._tfexample
    actual = utils.dict_to_tfexample(self._tfdict, store_shapes=True)
    self.assertProtoEquals(expected, actual)

  def test_parse_dict_to_tfexample(self):
    expected = self._tfdict
    # NOTE: TFExamples only store int64s so we must upcast bools and int32s.
    expected["bool"] = tf.cast(expected["bool"], tf.int64)
    expected["int32"] = tf.cast(expected["int32"], tf.int64)

    actual = tf.io.parse_single_example(
        serialized=tf.constant(self._tfexample.SerializeToString(), tf.string),
        features={
            "bool": tf.io.FixedLenFeature([2], dtype=tf.int64),
            "int32": tf.io.FixedLenFeature([1], dtype=tf.int64),
            "int64": tf.io.FixedLenFeature([1], dtype=tf.int64),
            "float": tf.io.FixedLenFeature([1], dtype=tf.float32),
            "string": tf.io.FixedLenFeature([1], dtype=tf.string),
            "2d_tensor": tf.io.FixedLenFeature([2, 2], dtype=tf.int64),
            "3d_tensor": tf.io.FixedLenFeature([2, 1, 3], dtype=tf.int64),
            "2d_ragged": tf.io.RaggedFeature(
                dtype=tf.int64,
                value_key="2d_ragged",
                partitions=(
                    tf.io.RaggedFeature.RowLengths(
                        utils.tfexample_ragged_length_key("2d_ragged", 0)
                    ),
                ),
            ),
            "3d_ragged": tf.io.RaggedFeature(
                dtype=tf.int64,
                value_key="3d_ragged",
                partitions=(
                    tf.io.RaggedFeature.RowLengths(
                        utils.tfexample_ragged_length_key("3d_ragged", 0)
                    ),
                    tf.io.RaggedFeature.RowLengths(
                        utils.tfexample_ragged_length_key("3d_ragged", 1)
                    ),
                ),
            ),
            "2d_sparse": tf.io.SparseFeature(
                value_key="2d_sparse",
                index_key=[
                    utils.tfexample_sparse_indices_key("2d_sparse", 0),
                    utils.tfexample_sparse_indices_key("2d_sparse", 1),
                ],
                size=[3, 4],
                dtype=tf.int64,
            ),
            "3d_sparse": tf.io.SparseFeature(
                value_key="3d_sparse",
                index_key=[
                    utils.tfexample_sparse_indices_key("3d_sparse", 0),
                    utils.tfexample_sparse_indices_key("3d_sparse", 1),
                    utils.tfexample_sparse_indices_key("3d_sparse", 2),
                ],
                size=[4, 5, 6],
                dtype=tf.int64,
            ),
        },
    )
    self.assertTensorDictEqual(expected, actual)

  def test_tfexample_to_dict(self):
    expected = dict(self._tfdict)
    # NOTE: TFExamples only store int64s so we must upcast bools and int32s.
    expected["bool"] = tf.cast(expected["bool"], tf.int64)
    expected["int32"] = tf.cast(expected["int32"], tf.int64)

    actual = utils.tfexample_to_dict(self._tfexample)
    self.assertTensorDictEqual(expected, actual)

  def test_dict_to_tfexample_legacy(self):
    features = {
        "inputs": "this is an input",
        "targets": "this is a target",
        "weight": 5.0,
        "idx1": np.array([1, 2], np.int32),
        "idx2": np.array([3, 4], np.int64),
        "is_correct": False,
        "2d_shape": np.arange(3).reshape((1, 3)),
        "3d_shape": np.arange(6).reshape((1, 2, 3)),
    }
    tfe = utils.dict_to_tfexample(features)

    self.assertLen(tfe.features.feature, len(features))
    self.assertEqual(
        tfe.features.feature["inputs"].bytes_list.value, [b"this is an input"]
    )
    self.assertEqual(
        tfe.features.feature["targets"].bytes_list.value, [b"this is a target"]
    )
    self.assertEqual(tfe.features.feature["weight"].float_list.value, [5.0])
    np.testing.assert_array_equal(
        tfe.features.feature["idx1"].int64_list.value,
        np.array([1, 2], np.int64),
    )
    np.testing.assert_array_equal(
        tfe.features.feature["idx2"].int64_list.value,
        np.array([3, 4], np.int64),
    )
    np.testing.assert_array_equal(
        tfe.features.feature["is_correct"].int64_list.value,
        np.array([0], np.int64),
    )
    np.testing.assert_array_equal(
        tfe.features.feature["2d_shape"].int64_list.value,
        np.arange(3).reshape((1, 3)).flatten(),
    )
    np.testing.assert_array_equal(
        tfe.features.feature["3d_shape"].int64_list.value,
        np.arange(6).reshape((1, 2, 3)).flatten(),
    )

  def test_flatten_dict(self):
    expected = {
        "key1/subkey1": tf.constant([1]),
        "key1/subkey2": tf.constant([2]),
        "key2/subkey3": tf.constant([3]),
        "key3": tf.constant([4]),
    }

    actual = utils.flatten_dict({
        "key1": {
            "subkey1": tf.constant([1]),
            "subkey2": tf.constant([2]),
        },
        "key2": {
            "subkey3": tf.constant([3]),
        },
        "key3": tf.constant([4]),
    })

    self.assertDictEqual(expected, actual)

  def test_unflatten_dict(self):
    expected = {
        "key1": {
            "subkey1": tf.constant([1]),
            "subkey2": tf.constant([2]),
        },
        "key2": {
            "subkey3": tf.constant([3]),
        },
        "key3": tf.constant([4]),
    }

    actual = utils.unflatten_dict({
        "key1/subkey1": tf.constant([1]),
        "key1/subkey2": tf.constant([2]),
        "key2/subkey3": tf.constant([3]),
        "key3": tf.constant([4]),
    })
    self.assertTensorDictEqual(expected, actual)

  def test_stateless_shuffle(self):
    value = np.arange(6)
    expected_output_1 = np.array([0, 3, 4, 2, 1, 5])
    expected_output_2 = np.array([3, 4, 0, 2, 5, 1])
    np.testing.assert_array_equal(
        utils.stateless_shuffle(value, (0, 1)), expected_output_1
    )
    np.testing.assert_array_equal(
        utils.stateless_shuffle(value.reshape((2, 3)), (0, 1)),
        expected_output_1.reshape((2, 3)),
    )
    np.testing.assert_array_equal(
        utils.stateless_shuffle(value, (2, 3)), expected_output_2
    )

  @parameterized.parameters(
      utils.map_over_dataset, utils.map_over_dataset(num_parallel_calls=2)
  )
  def test_map_over_dataset_as_decorator(self, decorator):
    @decorator
    def square(x):
      return x**2

    ds = square(tf.data.Dataset.range(4))
    self.assertEqual(list(ds.as_numpy_iterator()), [0, 1, 4, 9])

  def test_map_over_dataset_as_decorator_with_seeds(self):
    @utils.map_over_dataset(num_seeds=2)
    def square(x, seeds):
      del seeds
      return x**2

    ds = square(tf.data.Dataset.range(4))  # pylint: disable=no-value-for-parameter
    self.assertEqual(list(ds.as_numpy_iterator()), [0, 1, 4, 9])

  @parameterized.parameters({}, {"num_parallel_calls": 2})
  def test_map_over_dataset_as_function(self, **kwargs):
    def square(x):
      return x**2

    square = utils.map_over_dataset(square, **kwargs)
    ds = square(tf.data.Dataset.range(4))
    self.assertEqual(list(ds.as_numpy_iterator()), [0, 1, 4, 9])

  def test_map_over_dataset_as_function_with_seeds(self):
    def square(x, seeds):
      del seeds
      return x**2

    square = utils.map_over_dataset(square, num_seeds=2)
    ds = square(tf.data.Dataset.range(4))
    self.assertEqual(list(ds.as_numpy_iterator()), [0, 1, 4, 9])

  def test_map_over_dataset_as_partial_function(self):
    @functools.partial(utils.map_over_dataset, num_parallel_calls=2)
    def square(x):
      return x**2

    ds = square(tf.data.Dataset.range(4))
    self.assertEqual(list(ds.as_numpy_iterator()), [0, 1, 4, 9])

  def test_map_over_dataset_as_partial_function_with_seeds(self):
    @functools.partial(utils.map_over_dataset, num_seeds=2)
    def square(x, seeds):
      del seeds
      return x**2

    ds = square(tf.data.Dataset.range(4))  # pylint: disable=no-value-for-parameter
    self.assertEqual(list(ds.as_numpy_iterator()), [0, 1, 4, 9])

  # We disable no-value-for-parameter since the utils.map_over_dataset leads to
  # a false positive when seeds are provided.
  # pylint:disable=no-value-for-parameter

  def test_map_over_dataset_with_one_seed(self):
    inputs = tf.data.Dataset.range(2)

    tf.random.set_seed(None)
    utils._NEXT_MAP_SEED = 42

    @utils.map_over_dataset(num_seeds=1)
    def test_fn(x, seed):
      return x + seed

    expected = [
        np.array([2985944072, 3810604164]),
        np.array([4132877645, 4228622226]),
    ]
    for exp, act in zip(expected, test_fn(inputs).as_numpy_iterator()):
      np.testing.assert_array_equal(exp, act)

  def test_map_over_dataset_with_seeds(self):
    inputs = tf.data.Dataset.range(2)

    tf.random.set_seed(None)
    utils._NEXT_MAP_SEED = 42

    @utils.map_over_dataset(num_seeds=2)
    def test_fn(x, seeds):
      return x + seeds

    expected = [
        np.array([[2985944072, 3810604164], [64669036, 3548694723]]),
        np.array([[4132877645, 4228622226], [2495033825, 798765318]]),
    ]
    for exp, act in zip(expected, test_fn(inputs).as_numpy_iterator()):
      np.testing.assert_array_equal(exp, act)

  # pylint:enable=no-value-for-parameter

  def test_map_seed_manager(self):
    utils._NEXT_MAP_SEED = None
    self.assertIsNone(utils._NEXT_MAP_SEED)
    with utils.map_seed_manager(42):
      self.assertEqual(utils._NEXT_MAP_SEED, 42)
      with utils.map_seed_manager(410):
        self.assertEqual(utils._NEXT_MAP_SEED, 410)
        utils._NEXT_MAP_SEED += 10
        self.assertEqual(utils._NEXT_MAP_SEED, 420)
      utils._NEXT_MAP_SEED += 10
      self.assertEqual(utils._NEXT_MAP_SEED, 52)
    self.assertIsNone(utils._NEXT_MAP_SEED)

  def test_trim_and_pad_dataset(self):
    x = [
        {
            "inputs": [7, 8, 5, 6, 1],
            "targets": [[3, 0.5], [9, 0], [1, 2]],
            "idx": [0],
        },
        {
            "inputs": [8, 4, 9, 3, 5, 7, 9, 5],
            "targets": [[4, 1.2], [1, 1]],
            "idx": [1, 2],
        },
    ]
    ds = ds = tf.data.Dataset.from_generator(
        lambda: x,
        output_signature={
            "inputs": tf.TensorSpec([None], tf.int32),
            "targets": tf.TensorSpec([None, None], tf.float32),
            "idx": tf.TensorSpec([None], tf.int32),
        },
    )
    padded_ds = utils.trim_and_pad_dataset(
        ds, feature_lengths={"inputs": 7, "targets": 3}
    )
    expected = [
        {
            "inputs": [7, 8, 5, 6, 1, 0, 0],
            "targets": [[3, 0.5], [9, 0], [1, 2]],
            "idx": [0],
        },
        {
            "inputs": [8, 4, 9, 3, 5, 7, 9],
            "targets": [[4, 1.2], [1, 1], [0, 0]],
            "idx": [1, 2],
        },
    ]
    assert_dataset(
        padded_ds, expected, {"inputs": tf.int32, "targets": tf.float32}
    )

  def test_trim_and_pad_dataset_with_multirank_features(self):
    x = [
        {
            "inputs": [[[7, 8, 5, 6, 1]], [[1, 2, 3, 4, 5]]],
            "targets": [[3, 0.5], [9, 0], [1, 2]],
        },
        {
            "inputs": [[[8, 4, 9, 3, 5, 7, 9, 5]]],
            "targets": [[4, 1.2], [1, 1]],
        },
    ]
    ds = tf.data.Dataset.from_generator(
        lambda: x,
        output_signature={
            "inputs": tf.TensorSpec([None, None, None], tf.int32),
            "targets": tf.TensorSpec([None, None], tf.float32),
        },
    )
    padded_ds = utils.trim_and_pad_dataset(
        ds, feature_lengths={"inputs": [2, 1, 5], "targets": [3, 3]}
    )
    expected = [
        {
            "inputs": [[[7, 8, 5, 6, 1]], [[1, 2, 3, 4, 5]]],
            "targets": [[3, 0.5, 0], [9, 0, 0], [1, 2, 0]],
        },
        {
            "inputs": [[[8, 4, 9, 3, 5]], [[0, 0, 0, 0, 0]]],
            "targets": [[4, 1.2, 0], [1, 1, 0], [0, 0, 0]],
        },
    ]
    assert_dataset(
        padded_ds, expected, {"inputs": tf.int32, "targets": tf.float32}
    )

  _PACK_PARAMETERS = ({"use_custom_ops": False},)

  @parameterized.parameters(*_PACK_PARAMETERS)
  def test_trim_and_pack_dataset(self, use_custom_ops):
    x = [
        {"inputs": [7, 8, 5, 1], "targets": [3, 9, 1], "idx": [0]},
        {"inputs": [8, 4, 9, 3, 1], "targets": [4, 1], "idx": [1]},
    ]
    ds = create_default_dataset(x, feature_names=("inputs", "targets", "idx"))
    packed_ds = utils.trim_and_pack_dataset(
        ds,
        feature_lengths={"inputs": 10, "targets": 7},
        use_custom_ops=use_custom_ops,
    )

    expected = {
        "inputs": [7, 8, 5, 1, 8, 4, 9, 3, 1, 0],
        "inputs_segment_ids": [1, 1, 1, 1, 2, 2, 2, 2, 2, 0],
        "inputs_positions": [0, 1, 2, 3, 0, 1, 2, 3, 4, 0],
        "targets": [3, 9, 1, 4, 1, 0, 0],
        "targets_positions": [0, 1, 2, 0, 1, 0, 0],
        "targets_segment_ids": [1, 1, 1, 2, 2, 0, 0],
    }
    assert_dataset(
        packed_ds, expected, {"inputs": tf.int32, "targets": tf.int32}
    )


  @parameterized.parameters(*_PACK_PARAMETERS)
  def test_trim_and_pack_dataset_no_eos(self, use_custom_ops):
    x = [
        {"inputs": [7, 8, 5], "targets": [3, 9]},
        {"inputs": [8, 4, 9, 3], "targets": [4]},
    ]
    ds = create_default_dataset(x)
    packed_ds = utils.trim_and_pack_dataset(
        ds,
        feature_lengths={"inputs": 8, "targets": 5},
        use_custom_ops=use_custom_ops,
    )

    # Packing still works without the eos.
    expected = {
        "inputs": [7, 8, 5, 8, 4, 9, 3, 0],
        "inputs_segment_ids": [1, 1, 1, 2, 2, 2, 2, 0],
        "inputs_positions": [0, 1, 2, 0, 1, 2, 3, 0],
        "targets": [3, 9, 4, 0, 0],
        "targets_positions": [0, 1, 0, 0, 0],
        "targets_segment_ids": [1, 1, 2, 0, 0],
    }
    assert_dataset(
        packed_ds, expected, {"inputs": tf.int32, "targets": tf.int32}
    )

  @parameterized.parameters(*_PACK_PARAMETERS)
  def test_trim_and_pack_dataset_long_seq(self, use_custom_ops):
    x = [
        {"inputs": [7, 8, 5, 6, 9, 4, 1], "targets": [3, 9, 1]},
        {"inputs": [8, 4, 9, 3, 5, 7, 9, 1], "targets": [4, 1]},
    ]
    ds = create_default_dataset(x)
    packed_ds = utils.trim_and_pack_dataset(
        ds,
        feature_lengths={"inputs": 7, "targets": 3},
        use_custom_ops=use_custom_ops,
    )
    expected = [
        {
            "inputs": [7, 8, 5, 6, 9, 4, 1],
            "inputs_segment_ids": [1, 1, 1, 1, 1, 1, 1],
            "inputs_positions": [0, 1, 2, 3, 4, 5, 6],
            "targets": [3, 9, 1],
            "targets_positions": [0, 1, 2],
            "targets_segment_ids": [1, 1, 1],
        },
        {
            # EOS is trimmed
            "inputs": [8, 4, 9, 3, 5, 7, 9],
            "inputs_segment_ids": [1, 1, 1, 1, 1, 1, 1],
            "inputs_positions": [0, 1, 2, 3, 4, 5, 6],
            "targets": [4, 1, 0],
            "targets_positions": [0, 1, 0],
            "targets_segment_ids": [1, 1, 0],
        },
    ]
    assert_dataset(
        packed_ds, expected, {"inputs": tf.int32, "targets": tf.int32}
    )

  def test_autoregressive_inputs_unpacked(self):
    x = tf.constant([3, 8, 9, 5, 1, 0, 0])
    autoreg_inputs = utils.make_autoregressive_inputs(x)
    actual = self.evaluate(autoreg_inputs)
    expected = [0, 3, 8, 9, 5, 1, 0]
    self.assertAllEqual(actual, expected)
    self.assertEqual(actual.dtype, np.int32)

  def test_autoregressive_inputs_unpacked_2d(self):
    x = tf.constant([[3, 8, 1, 0, 0], [9, 5, 2, 0, 6]])
    autoreg_inputs = utils.make_autoregressive_inputs(x)
    actual = self.evaluate(autoreg_inputs)
    expected = [[0, 0, 0, 0, 0], [3, 8, 1, 0, 0]]
    self.assertAllEqual(actual, expected)
    self.assertEqual(actual.dtype, np.int32)

  def test_autoregressive_inputs_packed(self):
    x = tf.constant([3, 8, 1, 9, 1, 5, 4, 1, 0, 0])
    sequence_id = tf.constant([1, 1, 1, 2, 2, 3, 3, 3, 0, 0])
    autoreg_inputs = utils.make_autoregressive_inputs(
        x, sequence_id=sequence_id
    )
    actual = self.evaluate(autoreg_inputs)
    expected = [0, 3, 8, 0, 9, 0, 5, 4, 0, 0]
    self.assertAllEqual(actual, expected)
    self.assertEqual(actual.dtype, np.int32)

  def test_autoregressive_inputs_packed_2d(self):
    x = tf.constant([[3, 8, 1, 0, 0], [9, 5, 2, 0, 6]])
    sequence_id = tf.constant([1, 2])
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        (
            "Only 1-D sequences are supported with packing. "
            "Got a packed 2-D sequence."
        ),
    ):
      utils.make_autoregressive_inputs(x, sequence_id=sequence_id)

  def test_autoregressive_inputs_packed_non_eos(self):
    # In the correct input format, x[4] should have been 1 (EOS).
    x = tf.constant([3, 8, 1, 9, 6, 5, 4, 1, 0, 0])
    # sequence_id is correctly formatted.
    sequence_id = tf.constant([1, 1, 1, 2, 2, 3, 3, 3, 0, 0])
    autoreg_inputs = utils.make_autoregressive_inputs(
        x, sequence_id=sequence_id
    )
    actual = self.evaluate(autoreg_inputs)
    # The incorrect x[4] should not affect the output as long as the sequence_id
    # is correct.
    expected = [0, 3, 8, 0, 9, 0, 5, 4, 0, 0]
    self.assertAllEqual(actual, expected)
    self.assertEqual(actual.dtype, np.int32)

  def test_autoregressive_inputs_different_dtypes(self):
    x = tf.constant([3, 8, 1, 9.9, 1, 5, 4, 1, 0, 0])
    sequence_id = tf.constant([1, 1, 1, 2, 2, 3, 3, 3, 0, 0], tf.int32)
    autoreg_inputs = utils.make_autoregressive_inputs(
        x, sequence_id=sequence_id, output_dtype=tf.float32
    )
    actual = self.evaluate(autoreg_inputs)
    # The incorrect x[4] should not affect the output as long as the sequence_id
    # is correct.
    expected = [0, 3, 8, 0, 9.9, 0, 5, 4, 0, 0]
    self.assertAllClose(actual, expected)
    self.assertEqual(actual.dtype, np.float32)

  def test_shift_right_by_one(self):
    x = tf.constant([3, 8, 2, 9, 3, 5, 4, 1, 0, 0])
    shift_x = utils._shift_right_by_one(x)
    actual = self.evaluate(shift_x)
    expected = [0, 3, 8, 2, 9, 3, 5, 4, 1, 0]
    self.assertAllEqual(actual, expected)
    self.assertEqual(actual.dtype, np.int32)

  def test_shift_right_by_one_without_default_bos(self):
    x = tf.constant([3, 8, 2, 9, 3, 5, 4, 1, 0, 0])
    shift_x = utils._shift_right_by_one(x, bos_id=10)
    actual = self.evaluate(shift_x)
    expected = [10, 3, 8, 2, 9, 3, 5, 4, 1, 0]
    self.assertAllEqual(actual, expected)
    self.assertEqual(actual.dtype, np.int32)


class MixtureRateTest(test_utils.FakeTaskTest):

  def test_mixing_rate_num_examples(self):
    self.assertEqual(3.0, utils.mixing_rate_num_examples(self.cached_task))

    self.assertEqual(
        81.0, utils.mixing_rate_num_examples(self.cached_task, scale=27)
    )

    self.assertEqual(
        9.0,
        utils.mixing_rate_num_examples(
            self.cached_task, scale=27, temperature=2
        ),
    )

    self.assertEqual(
        2550.25,
        utils.mixing_rate_num_examples(
            self.cached_task, scale=27, temperature=0.5, maximum=50.5
        ),
    )

    # Test fallback.
    self.assertEqual(
        3.0,
        utils.mixing_rate_num_examples(
            self.cached_task, fallback_to_num_input_examples=False
        ),
    )
    with self.assertRaises(AssertionError):
      utils.mixing_rate_num_examples(
          self.uncached_task, fallback_to_num_input_examples=False
      )
    self.assertEqual(
        30.0,
        utils.mixing_rate_num_examples(
            self.uncached_task, fallback_to_num_input_examples=True
        ),
    )

  def test_mixing_rate_num_characters(self):
    task = mock.Mock(
        cache_dir="", get_cached_stats=lambda split: {"text_chars": 10}
    )
    rate = utils.mixing_rate_num_characters(task, temperature=0.7)
    self.assertAlmostEqual(rate, 26.8269579528)  # 10**(1 / 0.7)


def create_default_dataset(
    x: Sequence[Mapping[str, int]],
    feature_names: Sequence[str] = ("inputs", "targets"),
) -> tf.data.Dataset:
  output_types = {feature_name: tf.int32 for feature_name in feature_names}
  output_shapes = {feature_name: [None] for feature_name in feature_names}

  ds = tf.data.Dataset.from_generator(
      lambda: x, output_types=output_types, output_shapes=output_shapes
  )
  return ds




if __name__ == "__main__":
  absltest.main()



# Copyright 2023 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Dumps preprocessed tasks as TFRecord of tf.Examples.

Usage:
====================
seqio_cache_tasks \
--tasks=my_task_*,your_task \
--excluded_tasks=my_task_5 \
--output_cache_dir=/path/to/cache_dir \
--module_import=my.tasks \
--alsologtostderr

"""

import hashlib
import importlib
import os
import re

from absl import app
from absl import flags
from absl import logging

import apache_beam as beam
import seqio
from seqio import beam_utils
import tensorflow.compat.v2 as tf



# Significantly speeds up preprocessing in tf1.
tf.compat.v1.enable_eager_execution()

FLAGS = flags.FLAGS

flags.DEFINE_list(
    "tasks",
    None,
    (
        "Regexes matching task(s) to build a preprocessed dataset for. Will"
        " build all registered if not specified."
    ),
)
flags.DEFINE_list("excluded_tasks", None, "Regexes matching task(s) to skip.")
flags.DEFINE_string(
    "output_cache_dir", None, "The directory to output cached tasks to."
)
flags.DEFINE_list(
    "tasks_additional_cache_dirs",
    [],
    (
        "Additional directories to search for cached Tasks after checking the "
        "global caches and `output_cache_dir`."
    ),
)
flags.DEFINE_multi_string(
    "module_import",
    [],
    (
        "Modules to import. Use this, for example, to add new `Task`s to the "
        "global `TaskRegistry`."
    ),
)
flags.DEFINE_list(
    "pipeline_options",
    ["--runner=DirectRunner"],
    (
        "A comma-separated list of command line arguments to be used as options"
        " for the Beam Pipeline."
    ),
)
flags.DEFINE_boolean(
    "ignore_other_caches",
    False,
    (
        "If true, cache the task to output_cache_dir, ignoring cached data in "
        "other directories regardless of the value of `overwrite`."
    ),
)
flags.DEFINE_boolean(
    "overwrite",
    False,
    (
        "If true, overwrite the cached task even if it exists in the cached "
        "directories."
    ),
)
flags.DEFINE_integer(
    "min_shards",
    64,
    (
        "The minimum number of output shards to produce. Higher is better for "
        "improved online data shuffling during training."
    ),
)
flags.DEFINE_integer(
    "max_shards",
    0,
    (
        "The maximum number of output shards to produce. If the numbers of "
        "shards for the proprcossed tasks are larger than this value, they "
        "will be reduced to be max_shards."
    ),
)
flags.DEFINE_string(
    "tfds_data_dir",
    None,
    (
        "If set, this directory will be used to store datasets prepared by"
        " TensorFlow Datasets that are not available in the public TFDS GCS"
        " bucket. Note that this flag overrides the `tfds_data_dir` attribute"
        " of all `Task`s."
    ),
)
flags.DEFINE_integer(
    "base_seed",
    None,
    "The base seed used to reproducibly generate seeds for preprocessing.",
)

flags.DEFINE_boolean(
    "enable_char_counts",
    False,
    (
        "If set, counts characters in the preprocessed dataset and writes to"
        " stats.json. Detokenizes the dataset adds significant overhead. Enable"
        " with caution."
    ),
)



def _import_modules(modules):
  for module in modules:
    if module:
      importlib.import_module(module)


def run_pipeline(
    pipeline,
    task_names,
    cache_dir,
    excluded_tasks=None,
    modules_to_import=(),
    overwrite=False,
    ignore_other_caches=False,
    completed_file_contents="",
):
  """Run preprocess pipeline."""
  output_dirs = []
  # Includes all names by default.
  included_regex = re.compile(r"(%s\Z)" % r"\Z|".join(task_names or [".*"]))
  # Excludes only empty names by default.
  excluded_regex = re.compile(r"(%s\Z)" % r"\Z|".join(excluded_tasks or []))
  task_names = [
      t
      for t in seqio.TaskRegistry.names()
      if included_regex.match(t) and not excluded_regex.match(t)
  ]
  if not task_names:
    logging.warning(
        "No tasks have been selected from the task registry. "
        "Please make sure that the tasks you want cached exist in "
        "the task registry and haven't been excluded by the "
        "--excluded_tasks flag."
    )
  for task_name in task_names:
    task = seqio.TaskRegistry.get(task_name)
    if not task.supports_caching:
      logging.info(
          "Skipping task that does not support caching: '%s'", task.name
      )
      continue

    task_cache_dir = task.cache_dir
    output_dir = os.path.join(
        cache_dir, seqio.get_task_dir_from_name(task.name)
    )

    if ignore_other_caches:
      if task_cache_dir != output_dir:
        logging.warning(
            (
                "Task '%s' is already cached at %s but writing cached data to"
                " %s anyway because the --ignore_other_caches flag is set."
                " Please notethat the task and cached data may change and must"
                " be kept track of and re-cached manually by the user as"
                " needed."
            ),
            task.name,
            task_cache_dir,
            output_dir,
        )
    else:
      if task_cache_dir and not overwrite:
        logging.info(
            "Skipping task '%s', which exists in cache dir: %s",
            task.name,
            task_cache_dir,
        )
        continue

      if task_cache_dir and overwrite:
        if task_cache_dir == output_dir:
          # We were asked to overwrite the data, and the given directory that we
          # should generate the data in already has the data, then delete it.
          logging.warning(
              "Overwriting already cached data for task '%s' in cache_dir %s",
              task.name,
              output_dir,
          )
          tf.io.gfile.rmtree(output_dir)
        else:
          # Cannot overwrite, since cache_dir isn't same as task.cache_dir.
          logging.warning(
              (
                  "Not overwriting data in task.cache_dir since it is "
                  "different from cache_dir - %s vs %s"
              ),
              task.cache_dir,
              output_dir,
          )
          continue

    if not task.splits:
      logging.warning("Skipping task '%s' with no splits.", task.name)
      continue

    # Log this task to the terminal.
    logging.info("Caching task '%s' with splits: %s", task.name, task.splits)

    output_dirs.append(output_dir)
    completion_values = []

    if isinstance(task.source, seqio.FunctionDataSource):
      logging.warning(
          (
              "Task '%s' using FunctionDataSource cannot be distributed. If"
              " your dataset is large, you may be able to speed up"
              " preprocessing by sharding it and using a TfdsSource,"
              " TFExampleSource, or TextLineSource instead."
          ),
          task.name,
      )

    for split in task.splits:
      label = "%s_%s" % (task.name, split)
      if FLAGS.base_seed and FLAGS.base_seed != -1:
        # Create a unique, deterministic preprocessors seed for the task.
        task_uid = hashlib.md5(task.name.encode()).digest()
        task_preprocessor_seed = (
            int.from_bytes(task_uid, "little") + FLAGS.base_seed
        )
      else:
        task_preprocessor_seed = None
      pat = beam_utils.PreprocessTask(
          task,
          split,
          modules_to_import=modules_to_import,
          preprocessors_seed=task_preprocessor_seed,
          tfds_data_dir=FLAGS.tfds_data_dir,
      )
      if FLAGS.min_shards > 0:
        num_shards = max(len(pat.shards), FLAGS.min_shards)
      else:
        num_shards = len(pat.shards)

      if FLAGS.max_shards > 0:
        num_shards = min(num_shards, FLAGS.max_shards)

      examples = (
          pipeline
          | "%s_pat" % label >> pat
          # this reshuffle globally shuffles examples as a side-effect,
          # and should not be removed.
          | "%s_global_example_shuffle" % label >> beam.Reshuffle()
      )

      completion_values.append(
          examples
          | "%s_write_tfrecord" % label
          >> beam_utils.WriteExampleTfRecord(
              seqio.get_cached_tfrecord_prefix(output_dir, split),
              num_shards=num_shards,
          )
      )
      completion_values.append(
          examples
          | "%s_info" % label >> beam_utils.GetInfo(num_shards)
          | "%s_write_info" % label
          >> beam_utils.WriteJson(seqio.get_cached_info_path(output_dir, split))
      )
      completion_values.append(
          examples
          | "%s_stats" % label
          >> beam_utils.GetStats(
              task.output_features, enable_char_counts=FLAGS.enable_char_counts
          )
          | "%s_write_stats" % label
          >> beam_utils.WriteJson(
              seqio.get_cached_stats_path(output_dir, split)
          )
      )

    # After all splits for this task have completed, write COMPLETED files to
    # the task's output directory.
    _ = (
        completion_values
        | "%s_flatten_completion_values" % task.name >> beam.Flatten()
        | "%s_discard_completion_values" % task.name
        >> beam.Filter(lambda _: False)
        | "%s_write_completed_file" % task.name
        >> beam.io.textio.WriteToText(
            os.path.join(output_dir, "COMPLETED"),
            append_trailing_newlines=False,
            num_shards=1,
            shard_name_template="",
            header=completed_file_contents,
        )
    )


  return output_dirs


def main(_):
  flags.mark_flags_as_required(["output_cache_dir"])

  _import_modules(FLAGS.module_import)

  seqio.add_global_cache_dirs(
      [FLAGS.output_cache_dir] + FLAGS.tasks_additional_cache_dirs
  )

  if FLAGS.tfds_data_dir:
    seqio.set_tfds_data_dir_override(FLAGS.tfds_data_dir)

  pipeline_options = beam.options.pipeline_options.PipelineOptions(
      FLAGS.pipeline_options
  )
  with beam.Pipeline(options=pipeline_options) as pipeline:
    tf.io.gfile.makedirs(FLAGS.output_cache_dir)
    unused_output_dirs = run_pipeline(
        pipeline,
        FLAGS.tasks,
        FLAGS.output_cache_dir,
        FLAGS.excluded_tasks,
        FLAGS.module_import,
        FLAGS.overwrite,
        FLAGS.ignore_other_caches,
    )


def console_entry_point():
  app.run(main)


if __name__ == "__main__":
  console_entry_point()



# Copyright 2023 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Feature converters for common architectures.

In short, feature converters carry out additional data processing to the
tf.data.Dataset out of the Task API. They convert the features of the input
dataset into more descriptive features (e.g., "decoder_target_tokens" instead of
"targets") as well as pad and/or pack them. The features of the input dataset
are referred to as "task_features" because they are the output of the Task API.
Those of the output dataset are referred to as "model_features" as they are the
features directly fed to the model implementation.

We provide feature converters for the following three architectures:

  - encoder-decoder
  - decoder-only
  - encoder-only

Each of these feature converters inherit the base class FeatureConverter and
override two methods `_convert_features` and `get_model_feature_lengths` to
define how task features are mapped to the model features including the length
relationships. Other model architectures can be supported by subclassing the
FeatureConverter class in a similar manner.


Definition: standard_features

Throughout this module, we refer to the following 10 fields as
standard_features. Depending on the model architecture, a subset of them will
be returned by the feature converter.

  - encoder_input_tokens
  - encoder_target_tokens
  - encoder_loss_weights
  - encoder_positions
  - encoder_segment_ids
  - decoder_input_tokens
  - decoder_target_tokens
  - decoder_loss_weights
  - decoder_positions
  - decoder_segment_ids

  *_segment_ids and *_positions fields are only relevant for packed dataset.

  *_segment_ids is a tf.Tensor of integer which is aligned with
  *_input_tokens. Positive integers represent the sequence membership in
  the packed examples. 0 represents padding. For example, encoder_segment_ids =
  [1, 1, 2, 2, 2, 0] means that the first two positions belong to the first
  sequence, the next three to the second sequence and the last position is a
  padding.

  *_positions is a tf.Tensor of integer representing the position index in the
  original sequence before packing. For example, consider
  encoder_positions = [0, 1, 0, 1, 2, 0]. The first two tokens were the 0th and
  1st tokens of the first sequence and next three tokens are the 0th, 1st and
  2nd tokens of the second sequence before packing.

  *_loss_weights is used to indicate which positions should be used for the loss
  calculation.


Underlying assumptions

The feature converters implemented in this module assume the following about the
input dataset.

  - If EOS tokens are required, they are already appended in the input dataset.
  - The input dataset is not batched.
"""
import abc
import dataclasses
import functools
from typing import Mapping, Sequence
from absl import logging

from seqio import utils
import tensorflow.compat.v2 as tf

# TODO(hwchung): remove this.
# pointer for backward compatilbility.
autoregressive_inputs = utils.make_autoregressive_inputs


def _check_lengths(
    ds: tf.data.Dataset,
    expected_lengths: Mapping[str, int],
    sequence_axis_mapping: Mapping[str, int],
    strict: bool,
    error_label: str,
) -> tf.data.Dataset:
  """Check the length of each feature in `ds` against `expected_lengths`.

  There are two checking criteria controlled by `strict` arg.

  If strict = True,
  for each feature in ds, check len(feature) == expected_lengths[feature].

  If strict = False,
  for each feature in ds, check len(feature) <= expected_lengths[feature].

  Features of the input dataset may have [None] shape. The assertion is run at
  the graph execution time when the length is determined.

  Args:
    ds: a tf.data.Dataset to be checked.
    expected_lengths: a mapping from a feature name to an expected length.
    sequence_axis_mapping: a mapping from feature name to its sequence
      dimension.
    strict: if true, the length of each feature should exactly match the
      expected length whereas false condition allows the length to be less than
      or equal to the expected length.
    error_label: a label used to indicate the validation stage

  Returns:
    ds: the same dataset as but with the assertion ops attached.
  """

  def _check_length(feat, v):
    if feat not in expected_lengths:
      return v

    if strict:
      error_message = (
          f"Feature '{feat}' has length not equal to the expected length of "
          f"{expected_lengths[feat]} during {error_label} validation"
      )
      assertion_op = functools.partial(
          tf.debugging.assert_equal, message=error_message
      )
    else:
      error_message = (
          f"Feature '{feat}' has length not less than or equal to the expected "
          f"length of {expected_lengths[feat]} during {error_label} validation"
      )
      assertion_op = functools.partial(
          tf.debugging.assert_less_equal, message=error_message
      )

    expected_length = tf.constant(expected_lengths[feat], dtype=tf.int64)
    sequence_axis = sequence_axis_mapping[feat]
    if isinstance(v, tf.RaggedTensor):
      if strict:
        # For strict mode we require all sequence dim lengths be equal.
        # We multiply by [1] to (potentially) broadcast to 1d since non-ragged
        # dimensions will return scalar rather than vector row_lengths.
        lengths = v.row_lengths(sequence_axis) * tf.constant(
            [1], dtype=tf.int64
        )
        tf.debugging.assert_equal(
            len(tf.unique(lengths)[0]),
            1,
            "Strict length check requires all RaggedTensor dimensions to be"
            f" equal, got {v.row_lengths()}",
        )
      actual_length = v.bounding_shape(axis=sequence_axis, out_type=tf.int64)
    else:
      actual_length = tf.shape(v, out_type=tf.int64)[sequence_axis]
    assertion_op(actual_length, expected_length)
    return v

  ds = ds.map(
      lambda ex: {k: _check_length(k, v) for k, v in ex.items()},
      num_parallel_calls=tf.data.experimental.AUTOTUNE,
  )

  return ds


def non_padding_position(
    tensor: tf.Tensor, dtype: tf.dtypes.DType = tf.int32, pad_id: int = 0
) -> tf.Tensor:
  """Return a tensor with 1 on non-padding and 0 on padding positions."""
  return tf.cast(tf.not_equal(tensor, pad_id), dtype=dtype)


def _check_exact_match(
    expected_features: Sequence[str],
    actual_features: Sequence[str],
    expected_feature_source: str,
    actual_feature_source: str,
) -> None:
  """Check whether expected and actual features match one-to-one."""
  expected_features = set(expected_features)
  actual_features = set(actual_features)

  if expected_features != actual_features:
    if actual_features - expected_features:
      extra_features = actual_features - expected_features
      raise ValueError(
          f"The {actual_feature_source} contains extra features not specified "
          f"in the {expected_feature_source}: {extra_features}"
      )
    else:
      missing_features = expected_features - actual_features
      raise ValueError(
          f"The {actual_feature_source} is missing features specified "
          f"in the {expected_feature_source}: {missing_features}"
      )


class FeatureConverter(abc.ABC):
  """Abstract base class for feature converters.

  Subclasses of FeatureConverter are used to convert the tf.data.Dataset
  instance from the Task API to features that are passed to the
  model implementation. Note that Task API has an attribute "output_features",
  which is referred to as "task features" in the context of FeatureConverter.

  Typically the task features contain keys: "inputs" and "targets". The model
  features are constructed based on what is consumed by the model architecture.
  For custom model architectures that require additional model features, one
  needs to subclass FeatureConverter.

  This conversion is fully specified by

    1. defining the mapping of the features in `_convert_features` method and
    2. defining the relationship between sequence lengths of input and output
       features in `get_model_feature_lengths` which is a function of
       task_feature_lengths.

  Therefore, a subclass of FeatureConverter should override `_convert_features`
  and `get_model_feature_lengths` methods.

  The actual feature conversion is done in the `__call__` method, which
  wraps around the `_convert_features` method in order to provide useful checks
  and ensure compatibilities. See `_validate_dataset` and `__call__` methods
  for more details.

  Other notes:

    If pack = True, each feature in the task features should be packable,
    i.e., 1-dimensional.

    Subclasses must override TASK_FEATURES and MODEL_FEATURES. If packing is
    used, they must override PACKING_FEATURE_DTYPES as well. These are the
    packing-specific features such as "*_segment_ids".

  Attributes:
    pack: whether to pack the dataset.
    use_custom_packing_ops: whether to use custom ops for packing.
    apply_length_check: if True, it checks whether output feature lengths are
      less than the lengths given by `sequence_length`.
    bos_id: bos id for decoder inputs.
  """

  @dataclasses.dataclass(frozen=True)
  class FeatureSpec:
    """Rank and dtype specifications for features."""

    dtype: tf.dtypes.DType
    rank: int = 1
    sequence_dim: int = 0

  TASK_FEATURES: Mapping[str, "FeatureConverter.FeatureSpec"]
  MODEL_FEATURES: Mapping[str, "FeatureConverter.FeatureSpec"]
  PACKING_FEATURE_DTYPES: Mapping[str, tf.dtypes.DType]

  def __init__(
      self,
      pack: bool = True,
      use_custom_packing_ops: bool = False,
      apply_length_check: bool = True,
      bos_id: int = 0,
  ):
    self._pack = pack
    self._use_custom_packing_ops = use_custom_packing_ops
    self._apply_length_check = apply_length_check
    self._bos_id = bos_id

    if self.TASK_FEATURES is None:
      raise ValueError("TASK_FEATURES must be defined in the subclass.")

    if self.MODEL_FEATURES is None:
      raise ValueError("MODEL_FEATURES must be defined in the subclass.")

    if self.pack and self.PACKING_FEATURE_DTYPES is None:
      raise ValueError(
          "PACKING_FEATURE_DTYPES must be defined in the subclass if pack=True."
      )

  def _validate_dataset(
      self,
      ds: tf.data.Dataset,
      expected_features: Mapping[str, "FeatureConverter.FeatureSpec"],
      expected_lengths: Mapping[str, int],
      strict: bool,
      error_label: str,
  ) -> tf.data.Dataset:
    """Validate properties of the dataset, raising Exceptions if needed.

    This method is used to validate whether the input dataset is compatible
    with the desired specifications. In particular, the following aspects are
    checked.

    Each feature in `expected_features`
      - is also in `ds`
      - is also in expected_lengths
      - is compatible with the expected lengths

    The compatibility of the length is controlled by strict. If true, the length
    of each feature should exactly match the expected length whereas false
    condition allows the length to be less than or equal to the expected length.

    Args:
      ds: a tf.data.Dataset to be validated
      expected_features: expected features
      expected_lengths: a mapping from feature to its length
      strict: whether the lengths should be strictly equal or a length less than
        or equal to expected length is allowed.
      error_label: a label used to indicate the validation stage

    Returns:
      ds: the same dataset as but with the assertion ops attached.
    """
    element_spec = ds.element_spec
    for feat in expected_features:
      if feat not in element_spec:
        raise ValueError(
            "Dataset is missing an expected feature during "
            f"{error_label} validation: '{feat}'"
        )

      if expected_features[feat].dtype != element_spec[feat].dtype:
        actual_dtype = element_spec[feat].dtype.name
        raise ValueError(
            f"Dataset has incorrect type for feature '{feat}' during "
            f"{error_label} validation: Got {actual_dtype}, expected "
            f"{expected_features[feat].dtype.name}"
        )

      if expected_features[feat].rank != len(element_spec[feat].shape):
        actual_rank = len(element_spec[feat].shape)
        raise ValueError(
            f"Dataset has incorrect rank for feature '{feat}' during "
            f"{error_label} validation: "
            f"Got {actual_rank}, expected {expected_features[feat].rank}"
        )

    sequence_axis_mapping = {
        feat: expected_features[feat].sequence_dim for feat in expected_features
    }
    # Remove rank-0 features from expected lengths to bypass the length check.
    expected_lengths = {
        k: v
        for k, v in expected_lengths.items()
        if k in expected_features and expected_features[k].rank != 0
    }
    if self._apply_length_check:
      ds = _check_lengths(
          ds, expected_lengths, sequence_axis_mapping, strict, error_label
      )
    else:
      logging.info(
          "Length validation is skipped since `apply_length_check=False`"
      )
    return ds

  def __call__(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    r"""Convert the features of `ds` into output features.

    This method should not be overridden by subclasses.

    There are two conversion steps and five validation steps.

    Conversion 1: task features are converted to model features in
                  `_convert_features

    Conversion 2: task_feature_lengths are converted to model_feature_lengths in
                  `get_model_feature_lengths`

    Validation 1: verifies that the user input `task_feature_lengths` only
                  contains the required features.

    Validation 2: verifies whether the input dataset has same or more features,
                  same dtype, and length that is less than or equal compared to
                  input_ds.

    Validation 3: partially verifies the behavior of overridden
                  `get_model_feature_lengths`.

    Validation 4: check whether the output dataset has expected features (extra
                  features are allowed), dtype, rank and lengths (exact match).

    Validation 5: check one-to-one match between the output dataset and
                  `expected_dtypes`. Extra features are not allowed.

    The following diagram describes the validation and conversion processes. We
    treat features in the TASK_FEATURES and MODEL_FEATURES specified as class
    variables as the ground-truth. For validations 3, 4 and 5, we define
    `expected_dtypes`.

    There are 5 validation steps. features (<=) means that features of the
    variable on the left is a subset of those of the variable on the right. For
    example, validation 2 guarantees that TASK_FEATURES has features that are a
    subset of the features of input_ds. Validation 4 has length (==), which
    means that it ensures that each feature in MODEL_FEATURES has the same
    length as the corresponding feature in output_ds.

    Overall, these 5 validations ensures that the output_ds has the expected
    features with exact length, dtype and rank. Again, these validations assume
    that TASK_FEATURES and MODEL_FEATURES are correct.


                        Validation 1                     Validation 2
    task_feature_lengths <-----------> TASK_FEATURES <----------------> input_ds
    |                   features (==)                    features (<=)        |
    |                                                    dtype (==)           |
    |                                                    length (<=)          |
    |                                                    rank (==1)           |
    |                                                                         |
    |   Conversion 2                                           Conversion 1   |
    | get_model_feature_lengths                             _convert_features |
    |                                                                         |
    |                                              Validation 5               |
    |                                           <-------------------->        |
    |                                                 features (==)           |
    |                                                                         |
    \/                    Validation 3                    Validation 4        \/
    model_feature_lengths <-------> expected_dtypes <----------------> output_ds
                        features (==)                     features (<=)
                                                          dtype (==)
                                                          length (==)
                                                          rank (==1)

    Args:
      ds: a tf.data.Dataset to be validated
      task_feature_lengths: a mapping from a task feature to its length

    Returns:
      ds: the converted dataset.
    """
    # Validation 1
    _check_exact_match(
        expected_features=list(self.TASK_FEATURES),
        actual_features=list(task_feature_lengths),
        expected_feature_source="TASK_FEATURES",
        actual_feature_source="task_feature_lengths",
    )

    # Validation 2
    ds = self._validate_dataset(
        ds,
        expected_features=self.TASK_FEATURES,
        expected_lengths=task_feature_lengths,
        # Before pack/pad, check feature (of ds) length <= task feature length
        strict=False,
        error_label="input_validation",
    )

    # Conversion 1: implemented by subclass
    ds = self._convert_features(ds, task_feature_lengths)

    expected_features = dict(self.MODEL_FEATURES)
    if self.pack:
      for k, v in expected_features.items():
        # Packing requires rank 1.
        if v.rank != 1 and not self._use_custom_packing_ops:
          raise ValueError(
              "When packing is enabled, expected ranks must be 1 or "
              f"use_custom_packing_ops must be set. Got expected rank {v.rank} "
              f"for feature {k}."
          )
      for k, v in self.PACKING_FEATURE_DTYPES.items():
        expected_features[k] = FeatureConverter.FeatureSpec(rank=1, dtype=v)

    # Conversion 2: implemented by subclasses
    model_feature_lengths = self.get_model_feature_lengths(task_feature_lengths)

    # Validation 3
    _check_exact_match(
        expected_features=list(expected_features),
        actual_features=list(model_feature_lengths),
        expected_feature_source="model_feature_names",
        actual_feature_source="model_feature_lengths",
    )

    # Validation 4
    ds = self._validate_dataset(
        ds,
        expected_features=expected_features,
        expected_lengths=model_feature_lengths,
        # After pack/pad, check feature (of ds) length == model feature length
        strict=True,
        error_label="output_validation",
    )

    # Validation 5
    _check_exact_match(
        expected_features=list(expected_features),
        actual_features=list(ds.element_spec),
        expected_feature_source="model_feature_names",
        actual_feature_source="output_dataset",
    )
    return ds

  def _pack_or_pad(
      self, ds: tf.data.Dataset, packed_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """Trim/pad to packed_lengths and optionally pack the input dataset."""
    if self.pack:
      ds = utils.trim_and_pack_dataset(
          ds, packed_lengths, self._use_custom_packing_ops
      )
    else:
      ds = utils.trim_and_pad_dataset(ds, packed_lengths)
    return ds

  @abc.abstractmethod
  def _convert_features(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """Main feature conversion method to be overridden.."""
    raise NotImplementedError

  @abc.abstractmethod
  def get_model_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    """Define the length relationship between task and model features."""
    raise NotImplementedError

  @property
  def pack(self) -> bool:
    return self._pack

  @property
  def bos_id(self) -> int:
    return self._bos_id


class EncDecFeatureConverter(FeatureConverter):
  """Feature converter for an encoder-decoder architecture.

  The input dataset has "inputs" and "targets" field. These will be converted
  to a subset of standard features.

  To use packing, pass pack = True argument to the FeatureConverter's
  constructor. When packing is done, two additional fields are added for each of
  "inputs" and "targets" fields.

  Example for a packed dataset:

  The input dataset has two examples each with "inputs" and "targets".

  ds = [{"inputs": [7, 8, 5, 1], "targets": [3, 9, 1]},
        {"inputs": [8, 4, 9, 3, 1], "targets": [4, 1]}]

  task_feature_lengths = {"inputs": 10, "targets": 7}

  First, the `inputs` are packed together, padded to length 10 and assigned to
  "encoder_input_tokens" field. The `targets` are processed similarly.

  The "*_segment_id" fields are generated from the packing operation. For the
  explanation of these fields, see the module docstring.

  The "decoder_loss_weights" is a binary mask indicating where non-padding
  positions are, i.e., value of 1 indicates non-padding and 0 for padding. This
  class assumes that the loss is taken only on the decoder side.

  converted_ds = [{
       "encoder_input_tokens": [7, 8, 5, 1, 8, 4, 9, 3, 1, 0],
        "encoder_segment_ids": [1, 1, 1, 1, 2, 2, 2, 2, 2, 0],
          "encoder_positions": [0, 1, 2, 3, 0, 1, 2, 3, 4, 0],
      "decoder_target_tokens": [3, 9, 1, 4, 1, 0, 0],
       "decoder_input_tokens": [0, 3, 9, 0, 4, 0, 0],
       "decoder_loss_weights": [1, 1, 1, 1, 1, 0, 0],
        "decoder_segment_ids": [1, 1, 1, 2, 2, 0, 0],
          "decoder_positions": [0, 1, 2, 0, 1, 0, 0],
  }]

  Note that two examples are packed together into one example.
  """

  TASK_FEATURES = {
      "inputs": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "targets": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  MODEL_FEATURES = {
      "encoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_target_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_loss_weights": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  PACKING_FEATURE_DTYPES = {
      "encoder_segment_ids": tf.int32,
      "decoder_segment_ids": tf.int32,
      "encoder_positions": tf.int32,
      "decoder_positions": tf.int32,
  }

  def _convert_example(
      self, features: Mapping[str, tf.Tensor]
  ) -> Mapping[str, tf.Tensor]:
    """Convert a seq2seq example into an example with model features."""
    # targets_segment_id is present only for a packed dataset.
    decoder_input_tokens = utils.make_autoregressive_inputs(
        features["targets"],
        sequence_id=features.get("targets_segment_ids", None),
        bos_id=self.bos_id,
    )

    d = {
        "encoder_input_tokens": features["inputs"],
        "decoder_target_tokens": features["targets"],
        "decoder_input_tokens": decoder_input_tokens,
        # Loss is computed for all but the padding positions.
        "decoder_loss_weights": non_padding_position(features["targets"]),
    }

    if self.pack:
      d["encoder_segment_ids"] = features["inputs_segment_ids"]
      d["decoder_segment_ids"] = features["targets_segment_ids"]
      d["encoder_positions"] = features["inputs_positions"]
      d["decoder_positions"] = features["targets_positions"]

    return d

  def _convert_features(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """Convert the dataset to be fed to the encoder-decoder model.

    The conversion process involves two steps

    1. Each feature in the `task_feature_lengths` is trimmed/padded and
       optionally packed depending on the value of self.pack.
    2. "inputs" fields are mapped to the encoder input and "targets" are mapped
       to decoder input (after being shifted) and target.

    All the keys in the `task_feature_lengths` should be present in the input
    dataset, which may contain some extra features that are not in the
    `task_feature_lengths`. They will not be included in the output dataset.
    One common scenario is the "inputs_pretokenized" and "targets_pretokenized"
    fields.

    Args:
      ds: an input tf.data.Dataset to be converted.
      task_feature_lengths: a mapping from feature to its length.

    Returns:
      ds: the converted dataset.
    """
    ds = self._pack_or_pad(ds, task_feature_lengths)
    return ds.map(
        self._convert_example, num_parallel_calls=tf.data.experimental.AUTOTUNE
    )

  def get_model_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    """Define the length relationship between input and output features."""
    encoder_length = task_feature_lengths["inputs"]
    decoder_length = task_feature_lengths["targets"]

    model_feature_lengths = {
        "encoder_input_tokens": encoder_length,
        "decoder_target_tokens": decoder_length,
        "decoder_input_tokens": decoder_length,
        "decoder_loss_weights": decoder_length,
    }
    if self.pack:
      model_feature_lengths["encoder_segment_ids"] = encoder_length
      model_feature_lengths["decoder_segment_ids"] = decoder_length
      model_feature_lengths["encoder_positions"] = encoder_length
      model_feature_lengths["decoder_positions"] = decoder_length

    return model_feature_lengths


class PrePackedEncDecFeatureConverter(EncDecFeatureConverter):
  """Feature converter for encoder-decoder with pre-packed examples.

  The input dataset has "inputs", "targets", "inputs_segment_ids",
  "inputs_positions", "targets_segment_ids", and "targets_positions". These will
  be converted to a subset of standard features.

  Since this feature converter assumes the data is already pre-packed, setting
  'pack=True' is not allowed.
  """

  TASK_FEATURES = {
      "inputs": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "targets": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "inputs_positions": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "inputs_segment_ids": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "targets_positions": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "targets_segment_ids": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  MODEL_FEATURES = {
      "encoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_target_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_loss_weights": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "encoder_positions": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "encoder_segment_ids": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_positions": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_segment_ids": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }

  def __init__(self, **kwargs) -> None:
    super().__init__(**kwargs)
    if self.pack:
      raise ValueError(
          "'pack=True' is not allowed in PrePackedEncDecFeatureConverter."
      )

  def _convert_features(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """Convert the dataset to be fed to the encoder-decoder model.

    See "PrePackedEncDecFeatureConverter._convert.features".

    Args:
      ds: an input tf.data.Dataset to be converted.
      task_feature_lengths: a mapping from feature to its length.

    Returns:
      ds: the converted dataset.
    """

    def convert_example(
        features: Mapping[str, tf.Tensor]
    ) -> Mapping[str, tf.Tensor]:
      # targets_segment_id is present only for a packed dataset.
      decoder_input_tokens = utils.make_autoregressive_inputs(
          features["targets"],
          sequence_id=features.get("targets_segment_ids", None),
          bos_id=self.bos_id,
      )

      d = {
          "encoder_input_tokens": features["inputs"],
          "decoder_target_tokens": features["targets"],
          "decoder_input_tokens": decoder_input_tokens,
          # Loss is computed for all but the padding positions.
          "decoder_loss_weights": non_padding_position(features["targets"]),
          "encoder_segment_ids": features["inputs_segment_ids"],
          "decoder_segment_ids": features["targets_segment_ids"],
          "encoder_positions": features["inputs_positions"],
          "decoder_positions": features["targets_positions"],
      }
      return d

    ds = utils.trim_and_pad_dataset(ds, task_feature_lengths)

    return ds.map(
        convert_example, num_parallel_calls=tf.data.experimental.AUTOTUNE
    )

  def get_model_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    """Define the length relationship between input and output features."""
    encoder_length = task_feature_lengths["inputs"]
    decoder_length = task_feature_lengths["targets"]

    model_feature_lengths = {
        "encoder_input_tokens": encoder_length,
        "decoder_target_tokens": decoder_length,
        "decoder_input_tokens": decoder_length,
        "decoder_loss_weights": decoder_length,
        "encoder_segment_ids": encoder_length,
        "decoder_segment_ids": decoder_length,
        "encoder_positions": encoder_length,
        "decoder_positions": decoder_length,
    }
    return model_feature_lengths


class LMFeatureConverter(FeatureConverter):
  """Feature converter for a language model (decoder-only) architecture.

  The input dataset must have "targets" field only.

  One common usecase is to pre-train a decoder-only model with the standard
  language modeling objective (i.e., predict the next token given the previous
  ones) on a unlabeled text corpus which only has "targets". Then the
  pre-trained model can be fine-tuned on a supervised task, e.g., machine
  translation by concatenating "inputs" and "targets". For this use case,
  pre-train with LMFeatureConverter and fine-tune with PrefixLMFeatureConverter.

  Example: a packed dataset.

    ds = [{"targets": [3, 9, 1]}, {"targets": [4, 1]}]

    input_lengths = {"targets": 6}

    converted_ds = {
        "decoder_target_tokens": [3, 9, 1, 4, 1, 0],
         "decoder_input_tokens": [0, 3, 9, 0, 4, 0],
         "decoder_loss_weights": [1, 1, 1, 1, 1, 0],
            "decoder_positions": [0, 1, 2, 0, 1, 0],
          "decoder_segment_ids": [1, 1, 1, 2, 2, 0]
    }
  Note that two examples are packed together into one example.
  """

  TASK_FEATURES = {"targets": FeatureConverter.FeatureSpec(dtype=tf.int32)}
  MODEL_FEATURES = {
      "decoder_target_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_loss_weights": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  PACKING_FEATURE_DTYPES = {
      "decoder_segment_ids": tf.int32,
      "decoder_positions": tf.int32,
  }

  def _convert_example(
      self, features: Mapping[str, tf.Tensor]
  ) -> Mapping[str, tf.Tensor]:
    """Convert an LM example into an example with model features."""
    # targets_segment_id is present only for a packed dataset.
    decoder_input_tokens = utils.make_autoregressive_inputs(
        features["targets"],
        sequence_id=features.get("targets_segment_ids", None),
        bos_id=self.bos_id,
    )

    d = {
        "decoder_target_tokens": features["targets"],
        "decoder_input_tokens": decoder_input_tokens,
        "decoder_loss_weights": non_padding_position(features["targets"]),
    }

    if self.pack:
      d["decoder_segment_ids"] = features["targets_segment_ids"]
      d["decoder_positions"] = features["targets_positions"]

    return d

  def _convert_features(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """Convert the dataset to be fed to a language model."""
    ds = self._pack_or_pad(ds, task_feature_lengths)
    return ds.map(
        self._convert_example, num_parallel_calls=tf.data.experimental.AUTOTUNE
    )

  def get_model_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    """Define the length relationship between task and model features."""
    decoder_length = task_feature_lengths["targets"]
    model_feature_lengths = {
        "decoder_target_tokens": decoder_length,
        "decoder_input_tokens": decoder_length,
        "decoder_loss_weights": decoder_length,
    }
    if self.pack:
      model_feature_lengths["decoder_segment_ids"] = decoder_length
      model_feature_lengths["decoder_positions"] = decoder_length

    return model_feature_lengths


class PrefixLMFeatureConverter(LMFeatureConverter):
  """Feature converter for a prefix language model architecture.

  The input dataset must have both "inputs" and "targets" fields. For language
  modeling objective with "targets" only dataset, use LMFeatureConverter.

  A decoder is a network which autoregressively produces an output sequence. It
  can be used for an input dataset which has a notion of "inputs" as well as
  "targets", (e.g., machine translation) by concatenating them to form the new
  targets. See Raffel et al. (2020), https://arxiv.org/abs/1910.10683, Section
  3.2.1 for a more detailed take on this topic.

  In the Prefix LM architecture discussed in Raffel et al. (2020), the tokens
  from the "inputs" portion are applied a fully visible self attention whereas
  those from "targets" are applied the causal self attention. This makes the
  contextual representation of the tokens from "inputs" bidirectional.

  In order to provide this information, this class provides an additional
  feature "decoder_causal_attention" on top of the model features returned by
  LMFeatureConverter. "decoder_causal_attention" is a binary mask where a value
  of 1 represents that the corresponding input token to the decoder belongs to
  the "inputs" before concatenation. Note that this attention mask is optional.
  For a model that does not require this feature, e.g., a fully causal masking
  on the concatenated sequence, the attention mask can be simply ignored.

  Note that "decoder_causal_attention" includes one additional position to the
  right. This is the position where the final token of the "inputs" (often an
  EOS) is read and the first "targets" token is predicted. This follows
  mesh_tensorflow/transformer/transformer.py

  Since "inputs" and "targets" are concatenated to form the new targets for the
  decoder, we might want to compute the loss only on the tokens that belong to
  "targets" before concatenation. This behavior is controlled by
  "loss_on_targets_only" attribute, which is passed to the constructor. By
  default, it is set to True. The resulting "decoder_loss_weights" therefore
  zeros out "inputs" portion as well as the padding tokens while having 1's on
  the targets token.

  Example 1: a packed dataset
  ```
  ds = [{"inputs": [7, 8, 5, 1], "targets": [3, 9, 1]},
        {"inputs": [8, 4, 9, 3, 1], "targets": [4, 1]}]

  task_feature_lengths = {"inputs": 7, "targets": 8}

  converted_ds = {
      "decoder_target_tokens": [7, 8, 5, 1, 3, 9, 1, 8, 4, 9, 3, 1, 4, 1, 0],
       "decoder_input_tokens": [0, 7, 8, 5, 1, 3, 9, 0, 8, 4, 9, 3, 1, 4, 0],
       "decoder_loss_weights": [0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0],
          "decoder_positions": [0, 1, 2, 3, 4, 5, 6, 0, 1, 2, 3, 4, 5, 6, 0],
        "decoder_segment_ids": [1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 0],
   "decoder_causal_attention": [1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0]
  }
  ```

  Example 2: unpacked dataset with extra long "inputs" `task_feature_lengths`
  ```
  ds = [{"inputs": [9, 4, 6, 1], "targets": [3, 9, 1]}]

  task_feature_lengths = {"inputs": 10, "targets": 4}

  converted_ds = {
         "decoder_target_tokens": [9, 4, 6, 1, 3, 9, 1, 0, 0, 0, 0, 0, 0, 0],
          "decoder_input_tokens": [0, 9, 4, 6, 1, 3, 9, 1, 0, 0, 0, 0, 0, 0],
          "decoder_loss_weights": [0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
      "decoder_causal_attention": [1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
  }

  Note that if the inputs length specified in `task_feature_lengths` is longer
  than the actual example length, the padding tokens are added after
  concatenation.
  ```

  Attributes:
    loss_on_targets_only: whether to compute loss on tokens which belonged to
      "targets" before concatenation.
  """

  TASK_FEATURES = {
      "inputs": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "targets": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  MODEL_FEATURES = {
      "decoder_target_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_loss_weights": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_causal_attention": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  PACKING_FEATURE_DTYPES = {
      "decoder_segment_ids": tf.int32,
      "decoder_positions": tf.int32,
  }

  def __init__(self, loss_on_targets_only: bool = True, **kwargs) -> None:
    self._loss_on_targets_only = loss_on_targets_only
    super().__init__(**kwargs)

  def _convert_example(
      self, features: Mapping[str, tf.Tensor]
  ) -> Mapping[str, tf.Tensor]:
    """Convert a Prefix LM example into an example with model features.

    Example:
    ```

    Suppose the original dataset is

    ds = [{"inputs": [9, 4, 6, 1], "targets": [3, 9, 1]}]

    Then the input features to this method (after padding) are

    features = {
                   "targets" = [9, 4, 6, 1, 3, 9, 1, 0, 0]
              "inputs_width" = [4, 4, 4, 4, 4, 4, 4, 0, 0]
      "inputs_width_add_pos" = [5, 5, 5, 5, 5, 5, 5, 0, 0]
    }

    where "inputs_width" is length of "inputs" tiled across length dimension and
    "inputs_width_add_pos" is the same except that it has one additional
    position.

    First the parent class's _convert_example method is used to obtain the
    standard LM features. Then we compute "decoder_causal_attention". For an
    unpacked dataset, we need to define the "positions" feature. Then,
    `positions < inputs_width_add_pos` gives the decoder_causal_attention.

        "inputs_width_add_pos" = [5, 5, 5, 5, 5, 5, 5, 0, 0]
                   "positions" = [0, 1, 2, 3, 4, 5, 6, 7, 8]
                           <     ---------------------------
    "decoder_causal_attention" = [1, 1, 1, 1, 1, 0, 0, 0, 0]

    Then, we compute the loss weights, which requires isolating the "targets"
    position. Here we use "inputs_width" feature to filter out the "inputs"
    portion. `padding_mask` has 1's on inputs and targets and 0's on padding.
    Taking XOR filters out the targets portion.

          "inputs_width" = [4, 4, 4, 4, 4, 4, 4, 0, 0]
             "positions" = [0, 1, 2, 3, 4, 5, 6, 0, 0]
                     <     ---------------------------
                  inputs = [1, 1, 1, 1, 0, 0, 0, 0, 0]
            padding_mask = [1, 1, 1, 1, 1, 1, 1, 0, 0]
                    xor    ---------------------------
    decoder_loss_weights = [0, 0, 0, 0, 1, 1, 1, 0, 0]

    Note that decoder_loss_weights is computed by the LMFeatureConverter.
    ```

    Args:
      features: an input tf.data.Dataset to be converted.

    Returns:
      d: the converted features.
    """
    # First use the standard LM conversion.
    lm_features = super()._convert_example(features)

    # Initialize the return dictionary with the lm features.
    d = dict(lm_features)

    if self.pack:
      positions = features["targets_positions"]
    # Without packing, targets_positions field does not exist.
    else:
      positions = tf.range(tf.size(features["targets"]))

    inputs_width = features["inputs_width_add_pos"]
    # Binary mask where 1 represents a position in a non-causal attention region
    d["decoder_causal_attention"] = tf.cast(
        positions < inputs_width, dtype=features["targets"].dtype
    )

    # When computing the loss weights with self.loss_on_targets_only = True, we
    # use features["inputs_width"], which encodes the number of "inputs" tokens.
    if self.loss_on_targets_only:
      # 1's on inputs and 0's on targets and padding.
      inputs = positions < features["inputs_width"]

      # 1's on inputs and targets and 0's on padding.
      padding_mask = tf.cast(d["decoder_loss_weights"], dtype=tf.bool)

      # XOR picks targets only. See docstring for an example.
      d["decoder_loss_weights"] = tf.cast(
          tf.math.logical_xor(inputs, padding_mask),
          dtype=features["targets"].dtype,
      )

    return d

  def _concat_and_add_masks(
      self, features: Mapping[str, tf.Tensor]
  ) -> Mapping[str, tf.Tensor]:
    """Creates concatenated inputs and targets fields and adds masks."""
    inputs = features["inputs"]
    targets = features["targets"]
    # If the targets are empty, we add one padding target.
    targets = tf.cond(
        tf.size(targets) > 0,
        lambda: targets,
        lambda: tf.zeros(1, dtype="int32"),
    )

    # Width of the "inputs" portion in the concatenated sequence.
    width = tf.size(inputs)
    inputs_width = tf.fill([tf.size(inputs) + tf.size(targets)], width)

    # Width with an extra position to the right in the inputs mask. See
    # docstring for details.
    inputs_width_add_pos = tf.fill(
        [tf.size(inputs) + tf.size(targets)], width + 1
    )

    return {
        "targets": tf.concat([inputs, targets], axis=-1),
        "inputs_width": inputs_width,
        "inputs_width_add_pos": inputs_width_add_pos,
    }

  def _concat_task_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    concat_length = sum(task_feature_lengths.values())
    return {
        "targets": concat_length,
        "inputs_width": concat_length,
        "inputs_width_add_pos": concat_length,
    }

  def _convert_features(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """Convert the input dataset to an output dataset to be fed to the model.

    The "inputs" and "targets" are concatenated to form the new targets. In
    addition, the binary mask to distinguish "inputs" and "targets" token are
    concatenated as well.

    We define inputs_width to be a width (or a number of tokens) of "inputs" in
    the concatenated sequence. This method computes the width corresponding with
    and without additional position. Both of these are necessary
    `_convert_example`.

    Args:
      ds: an input tf.data.Dataset to be converted.
      task_feature_lengths: a mapping from task feature name to its length.

    Returns:
      ds: the converted dataset.
    """

    def swap_vals(t, old_val, new_val):
      return tf.where(tf.equal(t, old_val), tf.fill([tf.size(t)], new_val), t)

    def swap_inputs_width(ex, old_val, new_val):
      ex["inputs_width"] = swap_vals(ex["inputs_width"], old_val, new_val)
      return ex

    replace_0s = functools.partial(swap_inputs_width, old_val=0, new_val=-1)
    restore_0s = functools.partial(swap_inputs_width, old_val=-1, new_val=-0)

    ds = ds.map(
        self._concat_and_add_masks,
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )

    concat_task_feature_lengths = self._concat_task_feature_lengths(
        task_feature_lengths
    )

    ds = ds.map(replace_0s, num_parallel_calls=tf.data.experimental.AUTOTUNE)
    ds = self._pack_or_pad(ds, concat_task_feature_lengths)
    ds = ds.map(restore_0s, num_parallel_calls=tf.data.experimental.AUTOTUNE)

    return ds.map(
        self._convert_example, num_parallel_calls=tf.data.experimental.AUTOTUNE
    )

  def get_model_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    """Define the length relationship between task and model features."""
    decoder_length = sum(task_feature_lengths.values())
    concat_length = {"targets": decoder_length}
    lm_model_feature_lengths = super().get_model_feature_lengths(concat_length)
    model_feature_lengths = dict(lm_model_feature_lengths)
    model_feature_lengths["decoder_causal_attention"] = decoder_length
    return model_feature_lengths

  @property
  def loss_on_targets_only(self) -> bool:
    return self._loss_on_targets_only


class PrefixSuffixLMFeatureConverter(PrefixLMFeatureConverter):
  """Feature converter for a input + target + suffix language model.

  When "suffixes" is an empty list, it is identical as PrefixLMFeatureConverter.
  When "suffixes" is not empty, it merges "targets" and "suffixes" but
  computes the loss only over tokens from "suffixes".

  Example: a packed dataset
  ```
  ds = [{"inputs": [9, 4, 6], "targets": [3, 9], "suffixes": [2, 1]},
        {"inputs": [3, 2,], "targets": [4,], "suffixes": []}]

  task_feature_lengths = {"inputs": 7, "targets": 8}

  converted_ds = {
      "decoder_target_tokens": [9, 4, 6, 3, 9, 2, 1, 3, 2, 4, 0, 0, 0, 0, 0],
      "decoder_input_tokens": [0, 9, 4, 6, 3, 9, 2, 0, 3, 2, 0, 0, 0, 0, 0],
      "decoder_loss_weights": [0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0],
      "target_suffix_weights": [0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0],
      "decoder_positions": [0, 1, 2, 3, 4, 5, 6, 0, 1, 2, 0, 0, 0, 0, 0],
      "decoder_segment_ids": [1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 0, 0, 0, 0, 0],
      "decoder_causal_attention": [1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0],
  }
  ```
  """

  TASK_FEATURES = {
      "inputs": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "targets": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "suffixes": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  MODEL_FEATURES = {
      "decoder_target_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_loss_weights": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_causal_attention": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "target_suffix_weights": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  PACKING_FEATURE_DTYPES = {
      "decoder_segment_ids": tf.int32,
      "decoder_positions": tf.int32,
  }

  def __init__(self, loss_on_targets_only: bool = True, **kwargs) -> None:
    self._loss_on_targets_only = loss_on_targets_only
    super().__init__(**kwargs)

  def _convert_example(
      self, features: Mapping[str, tf.Tensor]
  ) -> Mapping[str, tf.Tensor]:
    """Convert a example into an example with model features."""
    # First use the standard LM conversion.
    lm_features = super()._convert_example(features)
    d = dict(lm_features)
    target_suffix_weights = tf.cast(
        tf.equal(features["target_suffix_weights"], 2),
        dtype=d["decoder_loss_weights"].dtype)
    d["target_suffix_weights"] = target_suffix_weights
    return d

  def _concat_and_add_masks(
      self, features: Mapping[str, tf.Tensor]
  ) -> Mapping[str, tf.Tensor]:
    """Creates concatenated inputs and targets fields and adds masks."""
    inputs = features["inputs"]
    targets = features["targets"]
    suffixes = features["suffixes"]
    target_suffixes = tf.concat([targets, suffixes], axis=0)
    # If the targets are empty, we add one padding target.
    target_suffixes = tf.cond(
        tf.size(target_suffixes) > 0,
        lambda: target_suffixes,
        lambda: tf.zeros(1, dtype="int32"),
    )

    # Width of the "inputs" portion in the concatenated sequence.
    width = tf.size(inputs)
    inputs_width = tf.fill([tf.size(inputs) + tf.size(target_suffixes)], width)

    # Width with an extra position to the right in the inputs mask. See
    # docstring for PrefixSuffixLMFeatureConverter class for details.
    inputs_width_add_pos = tf.fill(
        [tf.size(inputs) + tf.size(target_suffixes)], width + 1
    )

    target_weights_with_suffix = tf.concat(
        [tf.ones_like(inputs),
         tf.ones_like(targets),
         tf.fill([tf.size(suffixes),], 2)], axis=-1)
    target_weights_without_suffix = tf.concat(
        [tf.ones_like(inputs),
         tf.fill([tf.size(target_suffixes),], 2)], axis=-1)

    target_weights = tf.cond(
        tf.size(suffixes) > 0,
        lambda: target_weights_with_suffix,
        lambda: target_weights_without_suffix,
    )
    return {
        "targets": tf.concat([inputs, target_suffixes], axis=-1),
        "inputs_width": inputs_width,
        "inputs_width_add_pos": inputs_width_add_pos,
        "target_suffix_weights": target_weights
    }

  def get_model_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
      ) -> Mapping[str, int]:
    """Define the length relationship between task and model features."""
    decoder_length = sum(task_feature_lengths.values())
    concat_length = {"targets": decoder_length}
    lm_model_feature_lengths = super().get_model_feature_lengths(concat_length)
    model_feature_lengths = dict(lm_model_feature_lengths)
    model_feature_lengths["decoder_causal_attention"] = decoder_length
    model_feature_lengths["target_suffix_weights"] = decoder_length
    return model_feature_lengths

  def _concat_task_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    concat_length = sum(task_feature_lengths.values())
    return {
        "targets": concat_length,
        "inputs_width": concat_length,
        "inputs_width_add_pos": concat_length,
        "target_suffix_weights": concat_length,
    }


class DecoderFeatureConverter(FeatureConverter):
  """Wrapper of FeatureConverter that handles both LM and PrefixLM tasks.

  The converter to choose depends on the keys of `task_feature_lengths`.
  """

  TASK_FEATURES = {
      "targets": FeatureConverter.FeatureSpec(dtype=tf.int32),
      # Optional fields:
      #   "inputs": FeatureConverter.FeatureSpec(dtype=tf.int32) - Runs
      #     PrefixLMFeatureConverter if present
      #   "suffixes": FeatureConverter.FeatureSpec(dtype=tf.int32) - Runs
      #     PrefixSuffixLMFeatureConverter is present. Runs LMFeatureConverter
      #     if neither are present.
  }
  MODEL_FEATURES = {
      "decoder_target_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "decoder_loss_weights": FeatureConverter.FeatureSpec(dtype=tf.int32),
      # Only if `inputs` field is present:
      "decoder_causal_attention": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  PACKING_FEATURE_DTYPES = {
      "decoder_segment_ids": tf.int32,
      "decoder_positions": tf.int32,
  }

  def __init__(
      self,
      loss_on_targets_only: bool = True,
      pack: bool = True,
      use_custom_packing_ops: bool = False,
      apply_length_check: bool = True,
      bos_id: int = 0,
  ) -> None:
    self._loss_on_targets_only = loss_on_targets_only
    super().__init__(
        pack=pack,
        use_custom_packing_ops=use_custom_packing_ops,
        apply_length_check=apply_length_check,
        bos_id=bos_id,
    )
    self.prefixlm_feature_converter = PrefixLMFeatureConverter(
        loss_on_targets_only=loss_on_targets_only,
        pack=pack,
        use_custom_packing_ops=use_custom_packing_ops,
        apply_length_check=apply_length_check,
        bos_id=bos_id,
    )
    self.strictlm_feature_converter = LMFeatureConverter(
        pack=pack,
        use_custom_packing_ops=use_custom_packing_ops,
        apply_length_check=apply_length_check,
        bos_id=bos_id,
    )
    self.prefixsuffixlm_feature_converter = PrefixSuffixLMFeatureConverter(
        loss_on_targets_only=loss_on_targets_only,
        pack=pack,
        use_custom_packing_ops=use_custom_packing_ops,
        apply_length_check=apply_length_check,
        bos_id=bos_id,
    )

  def __call__(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    # NOTE: __call__ can safely be overridden here because it is delegated to
    # other FeatureConverters - invoking their __call__ and associated checks.
    if "suffixes" in task_feature_lengths:
      return self.prefixsuffixlm_feature_converter(ds, task_feature_lengths)
    if "inputs" in task_feature_lengths:
      return self.prefixlm_feature_converter(ds, task_feature_lengths)
    else:
      return self.strictlm_feature_converter(ds, task_feature_lengths)

  def _convert_features(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """DecoderFeatureConverter does not have this method."""
    raise Exception("DecoderFeaturerConverter does not have this method.")

  def get_model_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    """Define the length relationship between task and model features."""
    if "suffixes" in task_feature_lengths:
      model_feature_lengths = (
          self.prefixsuffixlm_feature_converter.get_model_feature_lengths(
              task_feature_lengths
          )
      )
    elif "inputs" in task_feature_lengths:
      model_feature_lengths = (
          self.prefixlm_feature_converter.get_model_feature_lengths(
              task_feature_lengths
          )
      )
    else:
      model_feature_lengths = (
          self.strictlm_feature_converter.get_model_feature_lengths(
              task_feature_lengths
          )
      )
    return model_feature_lengths


class EncoderFeatureConverter(FeatureConverter):
  """Feature converter for encoder-only achitecture such as BERT.

  The inputs and targets to the encoder are expected to be aligned.

  Just like BERT (Devlin et al. 2019, https://arxiv.org/abs/1810.04805), a
  sentinel token (e.g., [CLS]) is expected to be prepended to the inputs and
  targets sequences. This ensures that the model can be used for a
  classification task. For a packed dataset, each sequence has separate sentinel
  tokens. In terms of segment_id, the classification sentinel is considered as a
  part of the sequence to which it is appended.

  Example for a packed dataset:

    The input dataset has two examples each with algined "inputs" and "targets".

    Here assume that mask_id = 9 and cls_id = 8

    ds = [{"inputs": [8, 9, 9, 3, 4, 1], "targets": [8, 7, 4, 3, 4, 1]},
          {"inputs": [8, 3, 9, 1], "targets": [8, 3, 6, 1]}]

    task_feature_lengths = {"inputs": 11, "targets": 11}

    converted_ds = [{
         "encoder_input_tokens": [8, 9, 9, 3, 4, 1, 8, 3, 9, 1, 0],
        "encoder_target_tokens": [8, 7, 4, 3, 4, 1, 8, 3, 6, 1, 0],
          "encoder_segment_ids": [1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 0],
            "encoder_positions": [0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 0],
         "encoder_loss_weights": [0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0]
    }]

    Note that two examples are packed together into one example.

  Attributes:
    mask_id: an integer indicating the mask sentinel token. This id is used to
      find the positions where the loss is taken.
  """

  TASK_FEATURES = {
      "inputs": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "targets": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  MODEL_FEATURES = {
      "encoder_target_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "encoder_input_tokens": FeatureConverter.FeatureSpec(dtype=tf.int32),
      "encoder_loss_weights": FeatureConverter.FeatureSpec(dtype=tf.int32),
  }
  PACKING_FEATURE_DTYPES = {
      "encoder_segment_ids": tf.int32,
      "encoder_positions": tf.int32,
  }

  def __init__(self, mask_id: int, **kwargs):
    self._mask_id = mask_id
    super().__init__(**kwargs)

  def _convert_features(
      self, ds: tf.data.Dataset, input_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """Convert the input dataset to an output dataset to be fed to the model.

    The conversion process involves three steps

    1. Each feature in the `input_lengths` is packed.
    2. "inputs" fields are mapped to the encoder input and "targets" are mapped
       to encoder target. Loss is taken only on the masked positions just as in
       Masked Language Modeling objective.

    Args:
      ds: an input tf.data.Dataset to be converted.
      input_lengths: a mapping from a feature to its length

    Returns:
      ds: the converted dataset.
    """

    @utils.map_over_dataset
    def convert_example(
        features: Mapping[str, tf.Tensor]
    ) -> Mapping[str, tf.Tensor]:
      inputs = features["inputs"]
      d = {
          "encoder_input_tokens": inputs,
          "encoder_target_tokens": features["targets"],
          "encoder_loss_weights": tf.cast(
              tf.equal(inputs, self.mask_id), tf.int32
          ),
      }

      if self.pack:
        d["encoder_segment_ids"] = features["inputs_segment_ids"]
        d["encoder_positions"] = features["inputs_positions"]

      return d

    ds = self._pack_or_pad(ds, input_lengths)
    return convert_example(ds)

  def get_model_feature_lengths(
      self, task_feature_lengths: Mapping[str, int]
  ) -> Mapping[str, int]:
    """Define the length relationship between input and output features."""
    encoder_length = task_feature_lengths["inputs"]
    model_feature_lengths = {
        "encoder_target_tokens": encoder_length,
        "encoder_input_tokens": encoder_length,
        "encoder_loss_weights": encoder_length,
    }
    if self.pack:
      model_feature_lengths["encoder_segment_ids"] = encoder_length
      model_feature_lengths["encoder_positions"] = encoder_length

    return model_feature_lengths

  @property
  def mask_id(self):
    return self._mask_id


class PassThroughFeatureConverter(FeatureConverter):
  """This feature converter pass through the dataset without any processing."""

  def __init__(self, **unused_kwargs):  # pylint: disable=super-init-not-called
    pass

  def __call__(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    del task_feature_lengths
    return ds

  def _convert_features(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ):
    """This method is required to be overridden but unused."""
    pass

  def get_model_feature_lengths(self, task_feature_lengths: Mapping[str, int]):
    """This method is required to be overridden but unused."""
    pass


class PrePackedLMFeatureConverter(PassThroughFeatureConverter):
  """This feature converter fixes length and filters batch features."""

  BATCH_FEATURES = (
      "decoder_input_tokens",
      "decoder_loss_weights",
      "decoder_positions",
      "decoder_segment_ids",
      "decoder_target_tokens",
  )

  def _set_shape_and_filter(self, ex, task_feature_lengths):
    shaped_filtered_ex = {}
    for feature in self.BATCH_FEATURES:
      ex[feature].set_shape(shape=task_feature_lengths["targets"])
      shaped_filtered_ex[feature] = ex[feature]
    return shaped_filtered_ex

  def __call__(
      self, ds: tf.data.Dataset, task_feature_lengths: Mapping[str, int]
  ) -> tf.data.Dataset:
    """Returns input dataset filtered for BATCH_FEATURES with fixed lengths.

    Args:
      ds: Prepacked dataset containing decoder BATCH_FEATURES and potentially
        additional features to be filtered out.
      task_feature_lengths: Dictionary of feature lengths. Must include a
        targets key, used for all decoder features.
    Returns: A dataset filtered to BATCH_FEATURES and with fixed set length.
    """
    return ds.map(
        functools.partial(
            self._set_shape_and_filter,
            task_feature_lengths=task_feature_lengths,
        ),
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )


class PrePackedPrefixLMFeatureConverter(PrePackedLMFeatureConverter):
  """Prefix LM variant of PrePackedLMFeatureConverter.

  The pass through feature converter fixes feature lengths and filters batch
  features. For Prefix LM, the inputs and targets are combined into each
  feature.
  """

  BATCH_FEATURES = PrePackedLMFeatureConverter.BATCH_FEATURES + (
      "decoder_causal_attention",
  )

  def _set_shape_and_filter(self, ex, task_feature_lengths):
    shaped_filtered_ex = {}
    for feature in self.BATCH_FEATURES:
      ex[feature].set_shape(
          shape=task_feature_lengths["inputs"] + task_feature_lengths["targets"]
      )
      shaped_filtered_ex[feature] = ex[feature]
    return shaped_filtered_ex


class GrainFeatureConverter(FeatureConverter):
  """Feature converter for Grain data pipeline."""

  def get_grain_transforms(
      self, task_feature_lengths: Mapping[str, int], batch_size: int
  ):
    raise NotImplementedError(
        "Need to implement the `get_grain_transforms` method which returns "
        "grain.Transformations."
    )
